import math
from typing import List, Tuple, Union

import numpy as np
import torch
from torch import Tensor


def fast_fbank_lfr_cmvn(
    waveform: np.ndarray,
    cmvn: np.ndarray,
    lfr_m: int = 5,
    lfr_n: int = 1,
    n_mels: int = 80,
    window: str = "hamming",
    frame_length: int = 25,
    frame_shift: int = 10,
    dither: float = 0.0,
    **kwargs,
) -> Tuple[np.ndarray, np.ndarray]:
    waveform = torch.from_numpy(waveform)

    feats = fast_fbank(waveform, n_mels, window, frame_length, frame_shift, dither)

    feats = feats.numpy().astype(np.float32)
    feats_len = np.array(feats.shape[0]).astype(np.int32)

    # LFR
    LFR_inputs = []
    inputs = feats
    T = inputs.shape[0]
    T_lfr = int(np.ceil(T / lfr_n))
    left_padding = np.tile(inputs[0], ((lfr_m - 1) // 2, 1))
    inputs = np.vstack((left_padding, inputs))
    T = T + (lfr_m - 1) // 2
    for i in range(T_lfr):
        if lfr_m <= T - i * lfr_n:
            LFR_inputs.append((inputs[i * lfr_n : i * lfr_n + lfr_m]).reshape(1, -1))
        else:
            # process last LFR frame
            num_padding = lfr_m - (T - i * lfr_n)
            frame = inputs[i * lfr_n :].reshape(-1)
            for _ in range(num_padding):
                frame = np.hstack((frame, inputs[-1]))

            LFR_inputs.append(frame)
    LFR_outputs = np.vstack(LFR_inputs).astype(np.float32)

    # CMVN
    inputs = LFR_outputs
    frame, dim = inputs.shape
    means = np.tile(cmvn[0:1, :dim], (frame, 1))
    vars = np.tile(cmvn[1:2, :dim], (frame, 1))
    inputs = (inputs + means) * vars

    feats = inputs
    feats_len = np.array(feats.shape[0]).astype(np.int32)
    return feats, feats_len


def load_cmvn(
    file_path: str = "checkpoints/iic/speech_fsmn_vad_zh-cn-16k-common-pytorch/am.mvn",
    return_numpy: bool = False,
):
    with open(file_path, "r", encoding="utf-8") as f:
        lines = f.readlines()

    means_list = []
    vars_list = []
    for i in range(len(lines)):
        line_item = lines[i].split()
        if line_item[0] == "<AddShift>":
            line_item = lines[i + 1].split()
            if line_item[0] == "<LearnRateCoef>":
                add_shift_line = line_item[3 : (len(line_item) - 1)]
                means_list = list(add_shift_line)
                continue
        elif line_item[0] == "<Rescale>":
            line_item = lines[i + 1].split()
            if line_item[0] == "<LearnRateCoef>":
                rescale_line = line_item[3 : (len(line_item) - 1)]
                vars_list = list(rescale_line)
                continue

    if return_numpy:
        means = np.array(means_list).astype(np.float64)
        vars = np.array(vars_list).astype(np.float64)
        cmvn = np.array([means, vars])
        return cmvn
    else:
        means = torch.tensor(means_list).float()
        vars = torch.tensor(vars_list).float()
        cmvn = torch.stack([means, vars])
        return cmvn


def fast_fbank(
    waveform: torch.Tensor,
    n_mels: int = 80,
    frame_length: int = 25,
    frame_shift: int = 10,
    dither: float = 0.0,
    sample_rate: int = 16000,
    remove_dc_offset: bool = True,
    raw_energy: bool = True,
    energy_floor: float = 0.0,
    preemphasis_coefficient: float = 0.97,
    use_power: bool = True,
    low_freq: float = 20.0,
    high_freq: float = 0.0,
    vtln_low: float = 100.0,
    vtln_high: float = -500.0,
    vtln_warp: float = 1.0,
    use_log_fbank: bool = True,
    round_to_power_of_two: bool = True,
) -> torch.Tensor:
    """基于 torchaudio和torch compile的快速fbank特征提取，输入为torch.Tensor类型的音频波形。
    fbank： 分帧 -》 高斯噪声 -》 移除直流分量 -》 计算能量 -》 预加重 -》 加窗 -》 傅里叶变换 -》 mel滤波器 -》 取对数

    Args:
        waveform (torch.Tensor): 输入的音频波形，shape为[T]，T为音频长度
        n_mels (int, optional): mel滤波器个数. Defaults to 80.
        frame_length (int, optional): 帧长,单位ms. Defaults to 25.
        frame_shift (int, optional): 帧移，单位ms. Defaults to 10.
        dither (float, optional): dither参数. Defaults to 0.0.
        sample_rate (int, optional): 采样率. Defaults to 16000.
        remove_dc_offset (bool, optional): 是否移除直流分量. Defaults to True.
        raw_energy (bool, optional): 是否计算原始能量,如果是则在预加重之前计算能量. Defaults to True.
        energy_floor (float, optional): 能量的最小值. Defaults to 0.0.
        preemphasis_coefficient (float, optional): 预加重系数. Defaults to 0.97.
        use_power (bool, optional): 是否使用功率谱. Defaults to True.
        low_freq (float, optional): 最低频率. Defaults to 20.0.
        high_freq (float, optional): 最高频率. Defaults to 0.0.
        vtln_low (float, optional): vtln warp的低频. Defaults to 100.0.
        vtln_high (float, optional): vtln warp的高频. Defaults to -500.0.
        vtln_warp (float, optional): vtln warp的系数. Defaults to 1.0.
        use_log_fbank (bool, optional): 是否对fbank结果取对数. Defaults to True.
        round_to_power_of_two (bool, optional): 是否将帧长调整为2的幂次方. Defaults to True.

    Returns:
        torch.Tensor: fbank特征，shape为[T, n_mels]
    """
    dtype, device = waveform.dtype, waveform.device
    waveform = waveform * (1 << 15)
    num_samples = waveform.size(0)
    eps = torch.tensor(torch.finfo(torch.float).eps)

    # 分帧
    window_length = int(frame_length * 16000 / 1000)
    window_shift = int(frame_shift * 16000 / 1000)
    m = 1 + (num_samples - window_length) // window_shift
    size = (m, window_length)
    strides = (window_shift, 1)
    strided_input = waveform.as_strided(size, strides)

    if dither > 0:
        # 高斯噪声
        rand_gauss = torch.randn(strided_input.shape, device=device, dtype=dtype)
        strided_input = strided_input + dither * rand_gauss

    if remove_dc_offset:
        # 移除直流分量
        # strided_input -= strided_input.mean(dim=1, dtype=dtype).unsqueeze(1)
        row_means = torch.mean(strided_input, dim=1).unsqueeze(1)  # size (m, 1)
        strided_input = strided_input - row_means
        strided_input = strided_input.to(device=device, dtype=dtype)

    if raw_energy:
        # 计算原始能量
        signal_log_energy = torch.max(strided_input.pow(2).sum(dim=-1), eps).log()
        if energy_floor > 0:
            signal_log_energy = torch.max(
                signal_log_energy,
                torch.tensor(math.log(energy_floor), device=device, dtype=dtype),
            )

    # 预加重
    if preemphasis_coefficient != 0:
        # strided_input[i,j] -= preemphasis_coefficient * strided_input[i, max(0, j-1)] for all i,j
        offset_strided_input = torch.nn.functional.pad(
            strided_input.unsqueeze(0), (1, 0), mode="replicate"
        ).squeeze(0)  # size (m, window_size + 1)
        strided_input = (
            strided_input - preemphasis_coefficient * offset_strided_input[:, :-1]
        )

    # 加窗
    window_function = torch.hamming_window(
        window_length, periodic=False, alpha=0.54, beta=0.46, device=device, dtype=dtype
    ).unsqueeze(0)

    strided_input = strided_input * window_function

    if round_to_power_of_two:
        # 将帧长调整为2的幂次方
        padded_window_length = _next_power_of_2(window_length)
        if padded_window_length != window_length:
            padding_right = padded_window_length - window_length
            strided_input = torch.nn.functional.pad(
                strided_input.unsqueeze(0), (0, padding_right), mode="constant", value=0
            ).squeeze(0)

    if not raw_energy:
        # 计算能量
        signal_log_energy = torch.max(strided_input.pow(2).sum(1), eps).log()
        if energy_floor > 0:
            signal_log_energy = torch.max(
                signal_log_energy,
                torch.tensor(math.log(energy_floor), device=device, dtype=dtype),
            )

    # 傅里叶变换
    # size (m, padded_window_size // 2 + 1)
    spectrum = torch.fft.rfft(strided_input).abs()

    # 功率谱
    if use_power:
        spectrum = spectrum.pow(2.0)

    # mel滤波器
    mel_energies, _ = get_mel_banks(
        n_mels,
        padded_window_length,
        sample_rate,
        low_freq,
        high_freq,
        vtln_low,
        vtln_high,
        vtln_warp,
    )
    mel_energies = mel_energies.to(device=device, dtype=dtype)
    # pad right column with zeros and add dimension, size (num_mel_bins, padded_window_size // 2 + 1)
    mel_energies = torch.nn.functional.pad(
        mel_energies, (0, 1), mode="constant", value=0
    )
    # sum with mel fiterbanks over the power spectrum, size (m, num_mel_bins)
    mel_energies = torch.mm(spectrum, mel_energies.T)

    # 取对数
    if use_log_fbank:
        # avoid log of zero (which should be prevented anyway by dithering)
        mel_energies = torch.max(mel_energies, eps).log()

    return mel_energies


def get_mel_banks(
    num_bins: int,
    window_length_padded: int,
    sample_freq: float,
    low_freq: float,
    high_freq: float,
    vtln_low: float,
    vtln_high: float,
    vtln_warp_factor: float,
) -> Tuple[Tensor, Tensor]:
    """
    Returns:
        (Tensor, Tensor): The tuple consists of ``bins`` (which is
        melbank of size (``num_bins``, ``num_fft_bins``)) and ``center_freqs`` (which is
        center frequencies of bins of size (``num_bins``)).
    """
    assert num_bins > 3, "Must have at least 3 mel bins"
    assert window_length_padded % 2 == 0
    num_fft_bins = window_length_padded / 2
    nyquist = 0.5 * sample_freq

    if high_freq <= 0.0:
        high_freq += nyquist

    assert (
        (0.0 <= low_freq < nyquist)
        and (0.0 < high_freq <= nyquist)
        and (low_freq < high_freq)
    ), "Bad values in options: low-freq {} and high-freq {} vs. nyquist {}".format(
        low_freq, high_freq, nyquist
    )

    # fft-bin width [think of it as Nyquist-freq / half-window-length]
    fft_bin_width = sample_freq / window_length_padded
    mel_low_freq = mel_scale_scalar(low_freq)
    mel_high_freq = mel_scale_scalar(high_freq)

    # divide by num_bins+1 in next line because of end-effects where the bins
    # spread out to the sides.
    mel_freq_delta = (mel_high_freq - mel_low_freq) / (num_bins + 1)

    if vtln_high < 0.0:
        vtln_high += nyquist

    assert vtln_warp_factor == 1.0 or (
        (low_freq < vtln_low < high_freq)
        and (0.0 < vtln_high < high_freq)
        and (vtln_low < vtln_high)
    ), (
        "Bad values in options: vtln-low {} and vtln-high {}, versus "
        "low-freq {} and high-freq {}".format(vtln_low, vtln_high, low_freq, high_freq)
    )

    bin = torch.arange(num_bins).unsqueeze(1)
    left_mel = mel_low_freq + bin * mel_freq_delta  # size(num_bins, 1)
    center_mel = mel_low_freq + (bin + 1.0) * mel_freq_delta  # size(num_bins, 1)
    right_mel = mel_low_freq + (bin + 2.0) * mel_freq_delta  # size(num_bins, 1)

    if vtln_warp_factor != 1.0:
        left_mel = vtln_warp_mel_freq(
            vtln_low, vtln_high, low_freq, high_freq, vtln_warp_factor, left_mel
        )
        center_mel = vtln_warp_mel_freq(
            vtln_low, vtln_high, low_freq, high_freq, vtln_warp_factor, center_mel
        )
        right_mel = vtln_warp_mel_freq(
            vtln_low, vtln_high, low_freq, high_freq, vtln_warp_factor, right_mel
        )

    center_freqs = inverse_mel_scale(center_mel)  # size (num_bins)
    # size(1, num_fft_bins)
    mel = mel_scale(fft_bin_width * torch.arange(num_fft_bins)).unsqueeze(0)

    # size (num_bins, num_fft_bins)
    up_slope = (mel - left_mel) / (center_mel - left_mel)
    down_slope = (right_mel - mel) / (right_mel - center_mel)

    if vtln_warp_factor == 1.0:
        # left_mel < center_mel < right_mel so we can min the two slopes and clamp negative values
        bins = torch.max(torch.zeros(1), torch.min(up_slope, down_slope))
    else:
        # warping can move the order of left_mel, center_mel, right_mel anywhere
        bins = torch.zeros_like(up_slope)
        up_idx = torch.gt(mel, left_mel) & torch.le(
            mel, center_mel
        )  # left_mel < mel <= center_mel
        down_idx = torch.gt(mel, center_mel) & torch.lt(
            mel, right_mel
        )  # center_mel < mel < right_mel
        bins[up_idx] = up_slope[up_idx]
        bins[down_idx] = down_slope[down_idx]

    return bins, center_freqs


def mel_scale_scalar(freq: float) -> float:
    return 1127.0 * math.log(1.0 + freq / 700.0)


def mel_scale(freq: Tensor) -> Tensor:
    return 1127.0 * (1.0 + freq / 700.0).log()


def inverse_mel_scale_scalar(mel_freq: float) -> float:
    return 700.0 * (math.exp(mel_freq / 1127.0) - 1.0)


def inverse_mel_scale(mel_freq: Tensor) -> Tensor:
    return 700.0 * ((mel_freq / 1127.0).exp() - 1.0)


def vtln_warp_freq(
    vtln_low_cutoff: float,
    vtln_high_cutoff: float,
    low_freq: float,
    high_freq: float,
    vtln_warp_factor: float,
    freq: Tensor,
) -> Tensor:
    r"""This computes a VTLN warping function that is not the same as HTK's one,
    but has similar inputs (this function has the advantage of never producing
    empty bins).

    This function computes a warp function F(freq), defined between low_freq
    and high_freq inclusive, with the following properties:
        F(low_freq) == low_freq
        F(high_freq) == high_freq
    The function is continuous and piecewise linear with two inflection
        points.
    The lower inflection point (measured in terms of the unwarped
        frequency) is at frequency l, determined as described below.
    The higher inflection point is at a frequency h, determined as
        described below.
    If l <= f <= h, then F(f) = f/vtln_warp_factor.
    If the higher inflection point (measured in terms of the unwarped
        frequency) is at h, then max(h, F(h)) == vtln_high_cutoff.
        Since (by the last point) F(h) == h/vtln_warp_factor, then
        max(h, h/vtln_warp_factor) == vtln_high_cutoff, so
        h = vtln_high_cutoff / max(1, 1/vtln_warp_factor).
          = vtln_high_cutoff * min(1, vtln_warp_factor).
    If the lower inflection point (measured in terms of the unwarped
        frequency) is at l, then min(l, F(l)) == vtln_low_cutoff
        This implies that l = vtln_low_cutoff / min(1, 1/vtln_warp_factor)
                            = vtln_low_cutoff * max(1, vtln_warp_factor)
    Args:
        vtln_low_cutoff (float): Lower frequency cutoffs for VTLN
        vtln_high_cutoff (float): Upper frequency cutoffs for VTLN
        low_freq (float): Lower frequency cutoffs in mel computation
        high_freq (float): Upper frequency cutoffs in mel computation
        vtln_warp_factor (float): Vtln warp factor
        freq (Tensor): given frequency in Hz

    Returns:
        Tensor: Freq after vtln warp
    """
    assert vtln_low_cutoff > low_freq, (
        "be sure to set the vtln_low option higher than low_freq"
    )
    assert vtln_high_cutoff < high_freq, (
        "be sure to set the vtln_high option lower than high_freq [or negative]"
    )
    l = vtln_low_cutoff * max(1.0, vtln_warp_factor)  # noqa
    h = vtln_high_cutoff * min(1.0, vtln_warp_factor)
    scale = 1.0 / vtln_warp_factor
    Fl = scale * l  # F(l)
    Fh = scale * h  # F(h)
    assert l > low_freq and h < high_freq
    # slope of left part of the 3-piece linear function
    scale_left = (Fl - low_freq) / (l - low_freq)
    # [slope of center part is just "scale"]

    # slope of right part of the 3-piece linear function
    scale_right = (high_freq - Fh) / (high_freq - h)

    res = torch.empty_like(freq)

    outside_low_high_freq = torch.lt(freq, low_freq) | torch.gt(
        freq, high_freq
    )  # freq < low_freq || freq > high_freq
    before_l = torch.lt(freq, l)  # freq < l
    before_h = torch.lt(freq, h)  # freq < h
    after_h = torch.ge(freq, h)  # freq >= h

    # order of operations matter here (since there is overlapping frequency regions)
    res[after_h] = high_freq + scale_right * (freq[after_h] - high_freq)
    res[before_h] = scale * freq[before_h]
    res[before_l] = low_freq + scale_left * (freq[before_l] - low_freq)
    res[outside_low_high_freq] = freq[outside_low_high_freq]

    return res


def vtln_warp_mel_freq(
    vtln_low_cutoff: float,
    vtln_high_cutoff: float,
    low_freq,
    high_freq: float,
    vtln_warp_factor: float,
    mel_freq: Tensor,
) -> Tensor:
    r"""
    Args:
        vtln_low_cutoff (float): Lower frequency cutoffs for VTLN
        vtln_high_cutoff (float): Upper frequency cutoffs for VTLN
        low_freq (float): Lower frequency cutoffs in mel computation
        high_freq (float): Upper frequency cutoffs in mel computation
        vtln_warp_factor (float): Vtln warp factor
        mel_freq (Tensor): Given frequency in Mel

    Returns:
        Tensor: ``mel_freq`` after vtln warp
    """
    return mel_scale(
        vtln_warp_freq(
            vtln_low_cutoff,
            vtln_high_cutoff,
            low_freq,
            high_freq,
            vtln_warp_factor,
            inverse_mel_scale(mel_freq),
        )
    )


def _next_power_of_2(x: int) -> int:
    r"""Returns the smallest power of 2 that is greater than x"""
    return 1 if x == 0 else 2 ** (x - 1).bit_length()


@torch.compile
def fast_pad_feats(feats: List[np.array], max_feat_len: Union[int, np.array]):
    padded_feats = []
    for feat in feats:
        if feat.shape[0] == max_feat_len:
            padded_feats.append(feat)
        else:
            pad_width = ((0, max_feat_len - feat.shape[0]), (0, 0))
            padded_feat = np.pad(feat, pad_width, "constant", constant_values=0)
            padded_feats.append(padded_feat)
    return np.array(padded_feats).astype(np.float32)



import ctypes
from typing import List
from uuid import uuid4

from loguru import logger
import numpy as np

from .audio import AudioChunk, AudioChunkList
from .waveform import Waveform


class AudioBytesStream:
    """
    Buffer and chunk audio byte data into fixed-size audiochunk(frame).

    This class is designed to handle incoming audio data in bytes,
    buffering it and producing audio frames of a consistent size.
    It is mainly used to easily chunk big or too small audio frames
    into a fixed size, helping to avoid processing very small frames
    (which can be inefficient) and very large frames (which can cause
    latency or processing delays). By normalizing frame sizes, it
    facilitates consistent and efficient audio data processing.
    """

    def __init__(
        self,
        sample_rate: int,
        num_channels: int = 1,
        chunk_size_ms: int = 100,
    ) -> None:
        """
        Initialize an AudioByteStream instance.

        Parameters:
            sample_rate (int): The audio sample rate in Hz.
            num_channels (int): The number of audio channels.
            samples_per_channel (int, optional): The number of samples per channel in each frame.
                If None, defaults to `sample_rate // 10` (i.e., 100ms of audio data).

        The constructor sets up the internal buffer and calculates the size of each frame in bytes.
        The frame size is determined by the number of channels, samples per channel, and the size
        of each sample (assumed to be 16 bits or 2 bytes).
        """
        self._sample_rate = sample_rate
        self._num_channels = num_channels
        samples_per_channel = sample_rate // 1000 * chunk_size_ms

        self._bytes_per_frame = (
            num_channels * samples_per_channel * ctypes.sizeof(ctypes.c_int16)
        )
        self._buf = bytearray()
        self._id = uuid4().hex

    def push(self, data: bytes) -> List[AudioChunk]:
        """
        Add audio data to the buffer and retrieve fixed-size frames.

        Parameters:
            data (bytes): The incoming audio data to buffer.

        Returns:
            AudioChunkList: A list of `AudioFrame` objects, each containing
            a fixed-size chunk of audio data.
        """
        self._buf.extend(data)

        chunks = AudioChunkList()
        while len(self._buf) >= self._bytes_per_frame:
            frame_data = self._buf[: self._bytes_per_frame]
            self._buf = self._buf[self._bytes_per_frame :]
            frame_data = np.frombuffer(
                frame_data,
                dtype=np.int16,
            ).astype(np.float32)
            chunks.append(
                AudioChunk(
                    stream_id=self._id,
                    waveform=Waveform(data=frame_data, sample_rate=self._sample_rate),
                )
            )

        return chunks

    def flush(self) -> AudioChunkList:
        """
        Flush the buffer and retrieve any remaining audio data as a final audio chunk.
        """
        chunks = AudioChunkList()
        if len(self._buf) == 0:
            return chunks

        if len(self._buf) % (2 * self._num_channels) != 0:
            logger.warning("AudioByteStream: incomplete frame during flush, dropping")
            return chunks

        chunks.append(
            AudioChunk(
                stream_id=self._id,
                waveform=Waveform(data=self._buf, sample_rate=self._sample_rate),
                is_last=True,
            )
        )
        self._buf.clear()
        return chunks


AudioBytesBuffer = (
    AudioBytesStream  # AudioBytesStream is deprecated, use AudioBytesBuffer instead.
)



# Copyright 2026 Xiaohongshu. (Author: Kaituo Xu, Kai Huang)

import functools
import glob
import logging
import os
import time

import soundfile as sf
from textgrid import IntervalTier, TextGrid

logger = logging.getLogger(__name__)


def get_wav_info(args):
    """
    Returns:
        wavs: list of (uttid, wav_path)
    """
    base = lambda p: os.path.basename(p).replace(".wav", "")
    if args.wav_path:
        wavs = [(base(args.wav_path), args.wav_path)]
    elif args.wav_paths and len(args.wav_paths) >= 1:
        wavs = [(base(p), p) for p in sorted(args.wav_paths)]
    elif args.wav_scp:
        with open(args.wav_scp) as fin:
            wavs = [line.strip().split() for line in fin]
    elif args.wav_dir:
        wavs = glob.glob(f"{args.wav_dir}/**/*.wav", recursive=True)
        wavs = [(base(p), p) for p in sorted(wavs)]
    else:
        raise ValueError("Please provide valid wav info")
    logger.info(f"#wavs={len(wavs)}")
    return wavs


def write_textgrid(wav_path, wav_dur, event):
    textgrid_file = wav_path.replace(".wav", ".TextGrid")
    logger.info(f"Write {textgrid_file}")
    textgrid = TextGrid(maxTime=wav_dur)
    tier = IntervalTier(name="voice", maxTime=wav_dur)
    for start_s, end_s in event:
        if start_s == end_s:
            logger.warning(f"Write TG, skip start=end {start_s}")
            continue
        start_s = max(start_s, 0)
        end_s = min(end_s, wav_dur)
        tier.add(minTime=start_s, maxTime=end_s, mark="1")
    textgrid.append(tier)
    textgrid.write(textgrid_file)


def write_event_textgrid(wav_path, wav_dur, event2starts_ends_s):
    textgrid_file = wav_path.replace(".wav", ".TextGrid")
    logger.info(f"Write {textgrid_file}")
    textgrid = TextGrid(maxTime=wav_dur)
    for event, starts_ends_s in event2starts_ends_s.items():
        tier = IntervalTier(name=event, maxTime=wav_dur)
        for start_s, end_s in starts_ends_s:
            if start_s == end_s:
                logger.warning(f"Write TG, skip start=end {start_s}")
                continue
            start_s = max(start_s, 0)
            end_s = min(end_s, wav_dur)
            tier.add(minTime=start_s, maxTime=end_s, mark="1")
        textgrid.append(tier)
    textgrid.write(textgrid_file)


def split_and_save_segment(wav_path, timestamps, save_segment_dir):
    logger.info("Split & save segment")
    os.makedirs(save_segment_dir, exist_ok=True)
    wav_np, sample_rate = sf.read(wav_path, dtype="int16")
    for j, (start_s, end_s) in enumerate(timestamps):
        uttid = wav_path.split("/")[-1].replace(".wav", "")
        seg_id = f"{uttid}_{j}_{int(start_s * 1000)}_{int(end_s * 1000)}"
        seg_path = f"{save_segment_dir}/{seg_id}.wav"
        start, end = int(start_s * sample_rate), int(end_s * sample_rate)
        sf.write(seg_path, wav_np[start:end], samplerate=sample_rate)


def split_and_save_event_segment(wav_path, event2timestamps, save_segment_dir):
    logger.info("Split & save segment")
    os.makedirs(save_segment_dir, exist_ok=True)
    wav_np, sample_rate = sf.read(wav_path, dtype="int16")
    for event, timestamps in event2timestamps.items():
        for i, (start_s, end_s) in enumerate(timestamps):
            uttid = wav_path.split("/")[-1].replace(".wav", "")
            seg_id = f"{uttid}_{event}_{i}_{int(start_s * 1000)}_{int(end_s * 1000)}"
            seg_path = f"{save_segment_dir}/{seg_id}.wav"
            start, end = int(start_s * sample_rate), int(end_s * sample_rate)
            sf.write(seg_path, wav_np[start:end], samplerate=sample_rate)


def timeit(func):
    # dur must be last return value of func
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        start = time.time()
        r = func(*args, **kwargs)
        elapsed = time.time() - start
        dur = r[-1]
        rtf = elapsed / dur if dur else 0
        logger.info(
            f"RTF={round(rtf, 5)}, elapsed={round(elapsed * 1000, 2)}ms, dur={dur}s"
        )
        return r

    return wrapper



from fasr_vad_fsmn.models.stream_vad import FSMNForStreamVAD, FSMNForStreamVADOnnx
from fasr_vad_fsmn.models.vad import FSMNForVAD

__all__ = [
    "FSMNForVAD",
    "FSMNForStreamVAD",
    "FSMNForStreamVADOnnx",
]



from __future__ import annotations

from pathlib import Path
from typing import ClassVar, List, Literal

from fireredasr2.asr import FireRedAsr2, FireRedAsr2Config
import librosa
import numpy as np
import torch
from typing_extensions import Self

from fasr.config import registry
from fasr.data import AudioToken, AudioTokenList
from fasr.model import ASRModel


def _any_tensor_to_numpy(x: np.ndarray | torch.Tensor) -> np.ndarray:
    if isinstance(x, torch.Tensor):
        return x.detach().cpu().float().numpy()
    return np.asarray(x)


def _prepare_firered_wav(
    data: np.ndarray | torch.Tensor, sample_rate: int, target_sr: int = 16000
) -> tuple[int, np.ndarray]:
    x = _any_tensor_to_numpy(data)
    if x.ndim > 1:
        x = np.mean(x, axis=0)
    x = x.flatten()
    if np.issubdtype(x.dtype, np.floating):
        xf = x.astype(np.float32)
    else:
        xf = x.astype(np.float32) / 32768.0
    if sample_rate != target_sr:
        xf = librosa.resample(xf, orig_sr=sample_rate, target_sr=target_sr)
        sample_rate = target_sr
    pcm = (np.clip(xf, -1.0, 1.0) * 32767.0).astype(np.int16)
    return sample_rate, pcm


def _tokens_from_aed_result(
    row: dict, return_timestamp: bool, duration_ms: int
) -> AudioTokenList:
    tokens = AudioTokenList()
    text = row.get("text") or ""
    if return_timestamp and row.get("timestamp"):
        for item in row["timestamp"]:
            if len(item) < 3:
                continue
            tok, start_s, end_s = item[0], item[1], item[2]
            tokens.append(
                AudioToken(
                    text=str(tok),
                    start_ms=int(round(float(start_s) * 1000)),
                    end_ms=int(round(float(end_s) * 1000)),
                )
            )
        return tokens
    dur = max(duration_ms, 2)
    tokens.append(AudioToken(text=text, start_ms=1, end_ms=dur))
    return tokens


def _tokens_from_llm_result(row: dict, duration_ms: int) -> AudioTokenList:
    text = row.get("text") or ""
    dur = max(duration_ms, 2)
    return AudioTokenList([AudioToken(text=text, start_ms=1, end_ms=dur)])


def _resolve_use_gpu(device: str | None) -> bool:
    if device is not None:
        return device != "cpu"
    return torch.cuda.is_available()


class _FireRedASRBase(ASRModel):
    """共享 FireRedAsr2 批推理与结果组装。"""

    kind: ClassVar[Literal["aed", "llm"]]
    _model: FireRedAsr2 | None = None
    _firered_config: FireRedAsr2Config | None = None

    def transcribe(
        self,
        batch: List[np.ndarray | torch.Tensor],
        **kwargs,
    ) -> List[AudioTokenList]:
        if self._model is None:
            raise RuntimeError("Call from_checkpoint before transcribe")
        sample_rate = int(kwargs.get("sample_rate", 16000))
        uttids = [f"utt-{i}" for i in range(len(batch))]
        wav_inputs: list[tuple[int, np.ndarray]] = []
        duration_ms_list: list[int] = []
        for w in batch:
            sr, pcm = _prepare_firered_wav(w, sample_rate)
            dur_ms = int(round(len(pcm) / float(sr) * 1000.0))
            duration_ms_list.append(max(dur_ms, 1))
            wav_inputs.append((sr, pcm))
        rows = self._model.transcribe(uttids, wav_inputs)
        by_utt = {r["uttid"]: r for r in rows}
        out: list[AudioTokenList] = []
        assert self._firered_config is not None
        for i, duration_ms in enumerate(duration_ms_list):
            row = by_utt.get(uttids[i], {"text": "", "uttid": uttids[i]})
            if self.kind == "aed":
                ts = self._firered_config.return_timestamp
                out.append(_tokens_from_aed_result(row, ts, duration_ms))
            else:
                out.append(_tokens_from_llm_result(row, duration_ms))
        return out

    def get_config(self):
        raise NotImplementedError

    def load(self, save_dir, **kwargs):
        raise NotImplementedError

    def save(self, save_dir, **kwargs):
        raise NotImplementedError


@registry.asr_models.register("firered")
@registry.asr_models.register("firered_aed")
class FireRedAEDForASR(_FireRedASRBase):
    """FireRed ASR-AED（CTC + Transformer 解码）。"""

    checkpoint: str = "FireRedTeam/FireRedASR2-AED"
    endpoint: str = "modelscope"
    kind: ClassVar[Literal["aed", "llm"]] = "aed"

    def from_checkpoint(
        self,
        checkpoint_dir: str | Path | None = None,
        device: str | None = None,
        use_half: bool = True,
        beam_size: int = 3,
        nbest: int = 1,
        decode_max_len: int = 0,
        softmax_smoothing: float = 1.25,
        aed_length_penalty: float = 0.6,
        eos_penalty: float = 1.0,
        return_timestamp: bool = True,
        elm_dir: str = "",
        elm_weight: float = 0.0,
        **kwargs,
    ) -> Self:
        _ = kwargs
        if not checkpoint_dir:
            checkpoint_dir = self.download_checkpoint()
        checkpoint_dir = Path(checkpoint_dir)
        if not checkpoint_dir.exists():
            raise FileNotFoundError(f"Checkpoint not found: {checkpoint_dir}")
        cfg = FireRedAsr2Config(
            use_gpu=_resolve_use_gpu(device),
            use_half=use_half,
            beam_size=beam_size,
            nbest=nbest,
            decode_max_len=decode_max_len,
            softmax_smoothing=softmax_smoothing,
            aed_length_penalty=aed_length_penalty,
            eos_penalty=eos_penalty,
            return_timestamp=return_timestamp,
            decode_min_len=0,
            repetition_penalty=1.0,
            llm_length_penalty=0.0,
            temperature=1.0,
            elm_dir=elm_dir,
            elm_weight=elm_weight,
        )
        self._model = FireRedAsr2.from_pretrained("aed", str(checkpoint_dir), cfg)
        self._firered_config = cfg
        return self


@registry.asr_models.register("firered_llm")
class FireRedLLMForASR(_FireRedASRBase):
    """FireRed ASR-LLM（语音编码 + 大模型解码）。"""

    checkpoint: str = "FireRedTeam/FireRedASR2-LLM"
    endpoint: str = "modelscope"
    kind: ClassVar[Literal["aed", "llm"]] = "llm"

    def from_checkpoint(
        self,
        checkpoint_dir: str | Path | None = None,
        device: str | None = None,
        beam_size: int = 3,
        decode_max_len: int = 0,
        decode_min_len: int = 0,
        repetition_penalty: float = 1.2,
        llm_length_penalty: float = 0.0,
        temperature: float = 1.0,
        **kwargs,
    ) -> Self:
        _ = kwargs
        if not checkpoint_dir:
            checkpoint_dir = self.download_checkpoint()
        checkpoint_dir = Path(checkpoint_dir)
        if not checkpoint_dir.exists():
            raise FileNotFoundError(f"Checkpoint not found: {checkpoint_dir}")
        cfg = FireRedAsr2Config(
            use_gpu=_resolve_use_gpu(device),
            use_bfloat16=True,
            beam_size=beam_size,
            nbest=1,
            decode_max_len=decode_max_len,
            softmax_smoothing=1.25,
            aed_length_penalty=0.6,
            eos_penalty=1.0,
            return_timestamp=False,
            decode_min_len=decode_min_len,
            repetition_penalty=repetition_penalty,
            llm_length_penalty=llm_length_penalty,
            temperature=temperature,
            elm_dir="",
            elm_weight=0.0,
        )
        self._model = FireRedAsr2.from_pretrained("llm", str(checkpoint_dir), cfg)
        self._firered_config = cfg
        return self



# Copyright 2026 Xiaohongshu. (Author: Kaituo Xu, Kai Huang)

import os

import torch
import torch.nn as nn
import torch.nn.functional as F


class DetectModel(nn.Module):
    @classmethod
    def from_pretrained(cls, model_dir):
        model_path = os.path.join(model_dir, "model.pth.tar")
        package = torch.load(
            model_path, map_location=lambda storage, loc: storage, weights_only=False
        )
        model = cls(package["args"])
        model.load_state_dict(package["model_state_dict"], strict=True)
        model.eval()
        return model

    def __init__(self, args):
        super().__init__()
        self.dfsmn = DFSMN(
            args.idim,
            args.R,
            args.M,
            args.H,
            args.P,
            args.N1,
            args.S1,
            args.N2,
            args.S2,
            args.dropout,
        )
        self.out = torch.nn.Linear(args.H, args.odim)

    @torch.no_grad()
    def forward(self, feat, caches=None):
        # type: (Tensor, Optional[List[Tensor]]) -> Tuple[Tensor, List[Tensor]]
        x, new_caches = self.dfsmn(feat, caches=caches)
        logits = self.out(x)
        probs = torch.sigmoid(logits)
        return probs, new_caches


class DFSMN(nn.Module):
    def __init__(self, D, R, M, H, P, N1, S1, N2=0, S2=0, dropout=0.1):
        """
        DFSMN config: Rx[H-P(N1,N2,S1,S2)]-MxH
        Args:
            D: input dimension
            R: number of DFSMN blocks
            M: number of DNN layers
            H: hidden size
            P: projection size
            N1: lookback order
            S1: lookback stride
            N2: lookahead order
            S2: lookahead stride
        """
        super().__init__()
        # Components
        # 1st FSMN block connecting input layer, without skip connection
        self.fc1 = nn.Sequential(
            nn.Linear(D, H, bias=True), nn.ReLU(), nn.Dropout(dropout)
        )
        self.fc2 = nn.Sequential(
            nn.Linear(H, P, bias=True), nn.ReLU(), nn.Dropout(dropout)
        )
        self.fsmn1 = FSMN(P, N1, S1, N2, S2)
        # N-1 DFSMN blocks
        self.fsmns = nn.ModuleList(
            [DFSMNBlock(H, P, N1, S1, N2, S2, dropout) for _ in range(R - 1)]
        )
        # M DNN layers
        dnn = [nn.Linear(P, H, bias=True), nn.ReLU(), nn.Dropout(dropout)]
        for l in range(M - 1):
            dnn += [nn.Linear(H, H, bias=True), nn.ReLU(), nn.Dropout(dropout)]
        self.dnns = nn.Sequential(*dnn)

    def forward(self, inputs, input_lengths=None, caches=None):
        # type: (Tensor, Optional[Tensor], Optional[List[Tensor]]) -> Tuple[Tensor, List[Tensor]]
        """
        Args:
            inputs: [N, T, D], padded, T is sequence length, D is input dim
            mask: processing padding issue, masked position is 1
                tensor.masked_fill(mask, value) will fill elements of tensor with value where mask is one.
        Returns:
            output: [N, T, P]
        """
        if input_lengths is None:
            mask = None
        else:
            mask = get_mask_from_lengths(input_lengths)
        # 1st FSMN
        h = self.fc1(inputs)
        p = self.fc2(h)
        new_caches = []
        if caches is None:
            cache = None
        else:
            cache = caches[0]
        memory, new_cache = self.fsmn1(p, mask=mask, cache=cache)
        new_caches.append(new_cache)

        # R-1 FSMN
        i = 1
        for fsmn in self.fsmns:
            if caches is None:
                cache = None
            else:
                cache = caches[i]
            memory, new_cache = fsmn(memory, mask=mask, cache=cache)
            new_caches.append(new_cache)
            i += 1
        # M DNN
        output = self.dnns(memory)
        return output, new_caches


def get_mask_from_lengths(lengths):
    """Mask position is set to 1 for Tensor.masked_fill(mask, value)
    Args:
        lengths: (N, )
    Return:
        mask: (N, T)
    """
    N = lengths.size(0)
    T = torch.max(lengths).item()
    mask = torch.zeros(N, T).to(lengths.device)
    for i in range(N):
        mask[i, lengths[i] :] = 1
    return mask.to(torch.uint8)


class DFSMNBlock(nn.Module):
    def __init__(self, H, P, N1, S1, N2=0, S2=0, dropout=0.1):
        """
        DFSMNBlock = [input -> Affine+ReLU -> Affine -> vFSMN -> output]
                        |                                           ^
                        |-------------------------------------------|
                                 (skip connection)
        Args:
            H: hidden size
            P: projection size
            N1: lookback order
            S1: lookback stride
            N2: lookahead order
            S2: lookahead stride
        """
        super().__init__()
        # Hyper-parameter
        self.H, self.P, self.N1, self.S1, self.N2, self.S2 = H, P, N1, S1, N2, S2
        # Components
        # step1. \hat{P}^{l-1} -> H^{l}, nonlinear affine transform
        self.fc1 = nn.Sequential(
            nn.Linear(P, H, bias=True), nn.ReLU(), nn.Dropout(dropout)
        )
        # Step2. H^{l} -> P^{l}, linear affine transform
        self.fc2 = nn.Linear(H, P, bias=False)
        # Step3. P^{l}-> \hat{P}^{l}, fsmn layer
        self.fsmn = FSMN(P, N1, S1, N2, S2)

    def forward(self, inputs, mask=None, cache=None):
        # type: (Tensor, Optional[Tensor], Optional[Tensor]) -> Tuple[Tensor, Tensor]
        """
        Args:
            inputs: [N, T, P], padded, T is sequence length, P is projection size
            mask: processing padding issue, masked position is 1
                tensor.masked_fill(mask, value) will fill elements of tensor with value where mask is one.
        Returns:
            output: [N, T, P]
        """
        residual = inputs
        # step1. \hat{P}^{l-1} -> H^{l}, nonlinear affine transform
        h = self.fc1(inputs)
        # Step2. H^{l} -> P^{l}, linear affine transform
        p = self.fc2(h)
        # Step3. P^{l}-> \hat{P}^{l}, fsmn layer
        memory, new_cache = self.fsmn(p, mask=mask, cache=cache)
        # Step4. skip connection
        output = memory + residual
        return output, new_cache


class FSMN(nn.Module):
    def __init__(self, P, N1, S1, N2=0, S2=0):
        """
        Args:
            P: projection size
            N1: lookback order
            S1: lookback stride
            N2: lookahead order
            S2: lookahead stride
        """
        super().__init__()
        # Hyper-parameter
        assert N1 >= 1
        self.N1, self.S1, self.N2, self.S2 = N1, S1, N2, S2
        # Components
        # P^{l}-> \hat{P}^{l}
        self.lookback_padding = (N1 - 1) * S1
        self.lookback_filter = nn.Conv1d(
            in_channels=P,
            out_channels=P,
            kernel_size=N1,
            stride=1,
            padding=self.lookback_padding,
            dilation=S1,
            groups=P,
            bias=False,
        )
        if self.N2 > 0:
            self.lookahead_filter = nn.Conv1d(
                in_channels=P,
                out_channels=P,
                kernel_size=N2,
                stride=1,
                padding=(N2 - 1) * S2,
                dilation=S2,
                groups=P,
                bias=False,
            )
        else:
            self.lookahead_filter = nn.Identity()

    def forward(self, inputs, mask=None, cache=None):
        # type: (Tensor, Optional[Tensor], Optional[Tensor]) -> Tuple[Tensor, Tensor]
        """
        Args:
            inputs: [N, T, P], padded, T is sequence length, P is projection size
            mask: processing padding issue, masked position is 1
                tensor.masked_fill(mask, value) will fill elements of tensor with value where mask is one.
        Returns:
            memory: [N, T, P]
        """
        T = inputs.size(1)
        if mask is not None:
            mask = mask.unsqueeze(-1)  # [N, T, 1]
            inputs = inputs.masked_fill(mask, 0.0)

        inputs = inputs.permute((0, 2, 1)).contiguous()  # [N, T, P] -> [N, P, T]
        residual = inputs

        if cache is not None:
            inputs = torch.cat((cache, inputs), dim=2)  # (N, P, C+T)
        new_cache = inputs[:, :, -self.lookback_padding :]  # (N, P, Co)

        # P^{l}-> \hat{P}^{l}, fsmn layer
        lookback = self.lookback_filter(inputs)
        if self.N1 > 1:
            lookback = lookback[:, :, : -(self.N1 - 1) * self.S1]
            if cache is not None:
                start = cache.size(2)
                lookback = lookback[:, :, start:]
            memory = residual + lookback
        else:
            memory = residual + lookback

        if self.N2 > 0 and T > 1:
            lookahead = self.lookahead_filter(inputs)
            memory += F.pad(lookahead[:, :, self.N2 * self.S2 :], (0, self.S2))
        memory = memory.permute((0, 2, 1)).contiguous()  # [N, P, T] -> [N, T, P]

        if mask is not None:
            memory = memory.masked_fill(mask, 0.0)
        return memory, new_cache



from pathlib import Path
import re
from typing import List

from funasr import AutoModel
from typing_extensions import Self

from fasr.config import registry
from fasr.data import AudioSpan, AudioSpanList, AudioToken, AudioTokenList
from fasr.model import PuncModel


@registry.punc_models.register("ct_transformer")
class CTTransformerForPunc(PuncModel):
    checkpoint: str = "iic/punc_ct-transformer_zh-cn-common-vocab272727-pytorch"
    endpoint: str = "modelscope"

    ct_transformer: AutoModel | None = None

    def restore(self, text: str) -> AudioSpanList:
        result = self.ct_transformer.generate(input=text)[0]
        punc_array = result.get("punc_array", []).tolist()
        sents = AudioSpanList()
        sent_tokens = []
        all_tokens = AudioTokenList(
            [AudioToken(text=t) for t in self.tokenize(text=text)]
        )
        assert len(all_tokens) == len(punc_array), (
            f"{len(all_tokens)} != {len(punc_array)}"
        )
        for i, punc_res in enumerate(punc_array):
            all_tokens[i].follow = self.id_to_punc(punc_res)
            if punc_res == 1:
                sent_tokens.append(all_tokens[i])
            else:
                sent_tokens.append(all_tokens[i])
                sents.append(
                    AudioSpan(
                        tokens=AudioTokenList(docs=sent_tokens),
                    )
                )
                sent_tokens = []
        if len(sent_tokens) > 0:
            sents.append(
                AudioSpan(
                    tokens=AudioTokenList(docs=sent_tokens),
                )
            )
        return sents

    def id_to_punc(self, id: int):
        punc_list = []
        for pun in self.ct_transformer.model.punc_list:
            if pun == "_":
                pun = ""
            punc_list.append(pun)
        id2punc = {i: punc for i, punc in enumerate(punc_list)}
        return id2punc[id]

    def tokenize(self, text: str) -> List[str]:
        return split_words(text=text)

    def from_checkpoint(
        self,
        checkpoint_dir: str | Path | None = None,
        disable_update: bool = True,
        disable_log: bool = True,
        disable_pbar: bool = True,
        **kwargs,
    ) -> Self:
        if not checkpoint_dir:
            checkpoint_dir = self.download_checkpoint()
        checkpoint_dir = Path(checkpoint_dir)
        if not checkpoint_dir.exists():
            raise FileNotFoundError(f"Checkpoint not found: {checkpoint_dir}")
        model = AutoModel(
            model=checkpoint_dir,
            disable_update=disable_update,
            disable_log=disable_log,
            disable_pbar=disable_pbar,
            **kwargs,
        )
        self.ct_transformer = model
        return self

    def get_config(self):
        raise NotImplementedError

    def load(self, save_dir, **kwargs):
        raise NotImplementedError

    def save(self, save_dir, **kwargs):
        raise NotImplementedError


def split_to_mini_sentence(words: list, word_limit: int = 20):
    assert word_limit > 1
    if len(words) <= word_limit:
        return [words]
    sentences = []
    length = len(words)
    sentence_len = length // word_limit
    for i in range(sentence_len):
        sentences.append(words[i * word_limit : (i + 1) * word_limit])
    if length % word_limit > 0:
        sentences.append(words[sentence_len * word_limit :])
    return sentences


def split_words(text: str, jieba_usr_dict=None, **kwargs):
    if jieba_usr_dict:
        input_list = text.split()
        token_list_all = []
        langauge_list = []
        token_list_tmp = []
        language_flag = None
        for token in input_list:
            if isEnglish(token) and language_flag == "Chinese":
                token_list_all.append(token_list_tmp)
                langauge_list.append("Chinese")
                token_list_tmp = []
            elif not isEnglish(token) and language_flag == "English":
                token_list_all.append(token_list_tmp)
                langauge_list.append("English")
                token_list_tmp = []

            token_list_tmp.append(token)

            if isEnglish(token):
                language_flag = "English"
            else:
                language_flag = "Chinese"

        if token_list_tmp:
            token_list_all.append(token_list_tmp)
            langauge_list.append(language_flag)

        result_list = []
        for token_list_tmp, language_flag in zip(token_list_all, langauge_list):
            if language_flag == "English":
                result_list.extend(token_list_tmp)
            else:
                seg_list = jieba_usr_dict.cut(
                    join_chinese_and_english(token_list_tmp), HMM=False
                )
                result_list.extend(seg_list)

        return result_list

    else:
        words = []
        segs = text.split()
        for seg in segs:
            # There is no space in seg.
            current_word = ""
            for c in seg:
                if len(c.encode()) == 1:
                    # This is an ASCII char.
                    current_word += c
                else:
                    # This is a Chinese char.
                    if len(current_word) > 0:
                        words.append(current_word)
                        current_word = ""
                    words.append(c)
            if len(current_word) > 0:
                words.append(current_word)
        return words


def isEnglish(text: str):
    if re.search("^[a-zA-Z']+$", text):
        return True
    else:
        return False


def join_chinese_and_english(input_list):
    line = ""
    for token in input_list:
        if isEnglish(token):
            line = line + " " + token
        else:
            line = line + token

    line = line.strip()
    return line



# Copyright 2026 Xiaohongshu. (Author: Kaituo Xu, Kai Huang)

from dataclasses import dataclass
import logging
import os
from typing import List, Tuple, Union

import numpy as np
import torch

from .core.audio_feat import AudioFeat
from .core.constants import FRAME_LENGTH_SAMPLE, FRAME_PER_SECONDS
from .core.detect_model import DetectModel
from .core.stream_vad_postprocessor import StreamVadFrameResult, StreamVadPostprocessor

logger = logging.getLogger(__name__)


@dataclass
class FireRedStreamVadConfig:
    use_gpu: bool = False
    smooth_window_size: int = 5
    speech_threshold: float = 0.5
    pad_start_frame: int = 5
    min_speech_frame: int = 8
    max_speech_frame: int = 2000  # 20s
    min_silence_frame: int = 20
    chunk_max_frame: int = 30000  # 300s

    def __post_init__(self):
        if self.speech_threshold < 0 or self.speech_threshold > 1:
            raise ValueError("speech_threshold must be in [0, 1]")
        if self.min_speech_frame <= 0:
            raise ValueError("min_speech_frame must be positive")


class FireRedStreamVad:
    @classmethod
    def from_pretrained(cls, model_dir, config=FireRedStreamVadConfig()):
        # Feat
        cmvn_path = os.path.join(model_dir, "cmvn.ark")
        feat_extractor = AudioFeat(cmvn_path)

        # Load & Build Model
        vad_model = DetectModel.from_pretrained(model_dir)
        if config.use_gpu:
            vad_model.cuda()
        else:
            vad_model.cpu()

        # Build Postprocessor
        postprocessor = StreamVadPostprocessor(
            config.smooth_window_size,
            config.speech_threshold,
            config.pad_start_frame,
            config.min_speech_frame,
            config.max_speech_frame,
            config.min_silence_frame,
        )
        return cls(feat_extractor, vad_model, postprocessor, config)

    def __init__(self, audio_feat, vad_model, postprocessor, config):
        self.audio_feat = audio_feat
        self.vad_model = vad_model
        self.postprocessor = postprocessor
        self.config = config
        self.model_caches = None

    def reset(self):
        self.model_caches = None
        self.audio_feat.reset()
        self.postprocessor.reset()

    def detect_frame(self, audio_frame: np.ndarray) -> StreamVadFrameResult:
        if len(audio_frame) != FRAME_LENGTH_SAMPLE:
            raise ValueError(
                f"Expected {FRAME_LENGTH_SAMPLE} samples, got {len(audio_frame)}"
            )

        feat, dur = self.audio_feat.extract(audio_frame)
        if self.config.use_gpu:
            feat = feat.cuda()

        prob, self.model_caches = self.vad_model.forward(
            feat.unsqueeze(0), caches=self.model_caches
        )
        raw_prob = prob.cpu().squeeze().tolist()

        frame_result = self.postprocessor.process_one_frame(raw_prob)
        return frame_result

    def detect_chunk(self, audio_chunk: np.ndarray) -> List[StreamVadFrameResult]:
        feats, dur = self.audio_feat.extract(audio_chunk)
        if self.config.use_gpu:
            feats = feats.cuda()

        probs, self.model_caches = self.vad_model.forward(
            feats.unsqueeze(0), caches=self.model_caches
        )
        raw_probs = probs.cpu().squeeze().tolist()
        if isinstance(raw_probs, float):
            raw_probs = [raw_probs]

        chunk_results = []
        for t, raw_prob in enumerate(raw_probs):
            stream_vad_frame_result = self.postprocessor.process_one_frame(raw_prob)
            chunk_results.append(stream_vad_frame_result)
        return chunk_results

    def detect_full(
        self, audio: Union[str, np.ndarray]
    ) -> Tuple[List[StreamVadFrameResult], dict]:
        self.reset()
        feats, dur = self.audio_feat.extract(audio)
        if self.config.use_gpu:
            feats = feats.cuda()

        if feats.size(0) <= self.config.chunk_max_frame:
            probs, _ = self.vad_model.forward(feats.unsqueeze(0))
        else:
            logger.warning(
                f"Too long input, split every {self.config.chunk_max_frame} frames"
            )
            chunk_probs = []
            chunks = feats.split(self.config.chunk_max_frame, dim=0)
            for chunk in chunks:
                chunk_prob, _ = self.vad_model.forward(chunk.unsqueeze(0))
                chunk_probs.append(chunk_prob)
            probs = torch.cat(chunk_probs, dim=1)
            probs = probs.squeeze()  # (T,)
        raw_probs = probs.cpu().squeeze().tolist()  # (T,)
        if isinstance(raw_probs, float):
            raw_probs = [raw_probs]

        frame_results = []
        for t, raw_prob in enumerate(raw_probs):
            stream_vad_frame_result = self.postprocessor.process_one_frame(raw_prob)
            frame_results.append(stream_vad_frame_result)
        self.reset()

        # Format result
        timestamps = self.results_to_timestamps(frame_results)
        result = {"dur": round(dur, 3), "timestamps": timestamps}
        if isinstance(audio, str):
            result["wav_path"] = audio
        return frame_results, result

    def set_mode(self, mode: int = 0):
        if mode == 0:  # VERY PERMISSIVE
            self.config.speech_threshold = 0.3
            self.config.min_speech_frame = 8
            self.config.min_silence_frame = 20
        elif mode == 1:  # PERMISSIVE
            self.config.speech_threshold = 0.5
            self.config.min_speech_frame = 10
            self.config.min_silence_frame = 15
        elif mode == 2:  # AGGRESSIVE
            self.config.speech_threshold = 0.7
            self.config.min_speech_frame = 15
            self.config.min_silence_frame = 10
        elif mode == 3:  # VERY_AGGRESSIVE
            self.config.speech_threshold = 0.9
            self.config.min_speech_frame = 20
            self.config.min_silence_frame = 5
        self.postprocessor.speech_threshold = self.config.speech_threshold
        self.postprocessor.min_speech_frame = self.config.min_speech_frame
        self.postprocessor.min_silence_frame = self.config.min_silence_frame

    @classmethod
    def results_to_timestamps(cls, results):
        results = sorted(results, key=lambda r: r.frame_idx)
        # Get frame index (0-based)
        frame_timestamps = []
        start, end = -1, -1
        for r in results:
            if r.is_speech_start:
                if start != -1:
                    logger.warning("start should be -1")
                start = max(0, r.speech_start_frame - 1)
                end = -1
            elif r.is_speech_end:
                assert end == -1
                end = max(0, r.speech_end_frame - 1)
                frame_timestamps.append((start, end))
                start, end = -1, -1
        if start != -1:
            assert end == -1
            end = results[-1].frame_idx - 1
            frame_timestamps.append((start, end))
        # Convert to seconds
        timestamps = []
        for s, e in frame_timestamps:
            timestamps.append((s / FRAME_PER_SECONDS, e / FRAME_PER_SECONDS))
        return timestamps



from pathlib import Path
from typing import Dict, Union

import yaml


def read_yaml(yaml_path: Union[str, Path]) -> Dict:
    """Read yaml file.

    Args:
        yaml_path (Union[str, Path]): The path of the yaml file.

    Raises:
        FileExistsError: If the file does not exist.

    Returns:
        Dict: The data in the yaml file.
    """
    if not Path(yaml_path).exists():
        raise FileExistsError(f"The {yaml_path} does not exist.")

    with open(str(yaml_path), "rb") as f:
        data = yaml.load(f, Loader=yaml.Loader)
    return data



from typing import Literal

from pydantic import BaseModel


class AudioChunkResponse(BaseModel):
    text: str
    state: (
        Literal[
            "segment_start", "segment_end", "interim_transcript", "final_transcript"
        ]
        | None
    ) = None
    lang: str | None = None
    start_time: float | None = None
    end_time: float | None = None
    confidence: float | None = None


class TranscriptionResponse(BaseModel):
    code: Literal[0, -1] = 0
    msg: str = "success"
    data: AudioChunkResponse | None = None



import asyncio
from asyncio import Event, Queue
from pathlib import Path
import traceback
from typing import List, Literal
from urllib.parse import parse_qs

from fastapi import FastAPI, WebSocket, WebSocketDisconnect
from fastapi.middleware.cors import CORSMiddleware
from loguru import logger
from pydantic import BaseModel, ConfigDict, Field
import uvicorn

from fasr.config import registry
from fasr.data import AudioBytesStream, AudioChunk, AudioChunkList, AudioToken
from fasr.model import ASRModel, PuncModel, StreamASRModel, StreamVADModel

from .schema import AudioChunkResponse, TranscriptionResponse


class RealtimeASRService(BaseModel):
    model_config = ConfigDict(
        arbitrary_types_allowed=True, validate_assignment=True, extra="ignore"
    )
    # service
    host: str = Field("127.0.0.1", description="服务地址")
    port: int = Field(27000, description="服务端口")
    device: Literal["cpu", "cuda", "mps"] | None = Field(None, description="设备")

    # audio input
    sample_rate: int = Field(16000, description="音频采样率")
    bit_depth: int = Field(16, description="音频位深")
    channels: int = Field(1, description="音频通道数")

    # speech
    min_speech_duration_ms: int = Field(500, description="最小语音持续时间, 单位ms")

    # 1pass model
    asr_model_name: Literal[
        "stream_sensevoice", "stream_paraformer.torch", "stream_paraformer.onnx"
    ] = Field("stream_paraformer.torch", description="流式asr模型")
    asr_checkpoint_dir: str | Path | None = Field(
        None,
        description="asr模型路径",
    )
    asr_model: StreamASRModel = Field(None, description="asr模型")

    vad_model_name: Literal["stream_fsmn.torch", "stream_fsmn.onnx"] = Field(
        "stream_fsmn.torch", description="流式vad模型"
    )
    vad_model: StreamVADModel = Field(None, description="vad模型")
    vad_checkpoint_dir: str | Path | None = Field(
        None,
    )
    vad_chunk_size_ms: int = Field(100, description="音频分片大小")
    vad_end_silence_ms: int = Field(500, description="vad判定音频片段结束最大静音时间")
    vad_threshold: float = Field(
        0.6,
        description="vad判定阈值, 取值范围0~1，值越大，则需要更大声音量来触发vad，噪音环境下建议设置更高的阈值",
        le=1,
        ge=0,
    )
    vad_db_threshold: int = Field(
        -100,
        description="vad音量阈值,值越大，则需要更大音量来触发vad，噪音环境下建议设置更高的阈值",
    )

    # 2pass model
    offline_asr_model_name: Literal["sensevoice", "paraformer"] = Field(
        None, description="离线asr模型"
    )
    offline_asr_checkpoint_dir: str | Path | None = Field(
        None,
        description="asr模型路径",
    )
    offline_asr_model: ASRModel = Field(None, description="离线asr模型")

    punc_model_name: Literal["ct_transformer"] = Field(None, description="标点模型")
    punc_model: PuncModel = Field(None, description="标点模型")
    punc_checkpoint_dir: str | Path | None = Field(
        None,
    )

    def setup(self):
        app = FastAPI()
        app.add_middleware(
            CORSMiddleware,
            allow_origins=["*"],
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
        )
        logger.info(
            f"Start online ASR Service on {self.host}:{self.port}, device: {self.device}"
        )
        logger.info(
            f"ASR Model: {self.asr_model_name}, VAD Model: {self.vad_model_name}, Punc Model: {self.punc_model_name or 'None'}, Offline ASR Model: {self.offline_asr_model_name or 'None'}"
        )
        logger.info(
            f"VAD Config: chunk_size_ms: {self.vad_chunk_size_ms}, end_silence_ms: {self.vad_end_silence_ms}, threshold: {self.vad_threshold}, db_threshold: {self.vad_db_threshold}"
        )

        self.asr_model: StreamASRModel = registry.stream_asr_models.get(
            self.asr_model_name
        )()
        self.asr_model.from_checkpoint(
            checkpoint_dir=self.asr_checkpoint_dir,
            device=self.device,
        )
        self.vad_model: StreamVADModel = registry.stream_vad_models.get(
            self.vad_model_name
        )(
            chunk_size_ms=self.vad_chunk_size_ms,
            max_end_silence_time=self.vad_end_silence_ms,
            speech_noise_thres=self.vad_threshold,
            db_threshold=self.vad_db_threshold,
        )
        self.vad_model.from_checkpoint(
            checkpoint_dir=self.vad_checkpoint_dir,
            device=self.device,
        )
        if self.offline_asr_model_name is not None:
            self.offline_asr_model: ASRModel = registry.asr_models.get(
                self.offline_asr_model_name
            )()
            self.offline_asr_model.from_checkpoint(
                checkpoint_dir=self.offline_asr_checkpoint_dir,
                device=self.device,
            )
        if self.punc_model_name is not None:
            self.punc_model: PuncModel = registry.punc_models.get(
                self.punc_model_name
            )()
            self.punc_model.from_checkpoint(
                checkpoint_dir=self.punc_checkpoint_dir,
                device=self.device,
            )

        @app.websocket("/asr/realtime")
        async def transcribe(ws: WebSocket):
            try:
                # 解析请求参数
                await ws.accept()
                query_params = parse_qs(ws.scope["query_string"].decode())
                itn = query_params.get("itn", ["false"])[0].lower() == "true"
                model = query_params.get("model", ["paraformer"])[0].lower()
                chunk_size = int(self.vad_chunk_size_ms * self.sample_rate / 1000)
                logger.info(f"itn: {itn}, chunk_size: {chunk_size}, model: {model}")
                span_queue = Queue()
                bytes_buffer = AudioBytesStream(
                    sample_rate=self.sample_rate, chunk_size_ms=self.vad_chunk_size_ms
                )
                vad_is_done = Event()
                tasks = []
                tasks.append(
                    asyncio.create_task(
                        self.vad_task(
                            ws,
                            span_queue=span_queue,
                            bytes_buffer=bytes_buffer,
                            vad_done_event=vad_is_done,
                        )
                    )
                )
                tasks.append(
                    asyncio.create_task(
                        self.asr_task(
                            ws=ws, span_queue=span_queue, vad_is_done=vad_is_done
                        )
                    )
                )
                _done, _ = await asyncio.wait(
                    tasks, return_when=asyncio.FIRST_COMPLETED
                )
            except WebSocketDisconnect:
                logger.info("WebSocket disconnected")
            except Exception as e:
                logger.error(
                    f"Unexpected error: {e}\nCall stack:\n{traceback.format_exc()}"
                )
                await ws.close()
            finally:
                self.vad_model.remove_state(bytes_buffer._id)
                self.asr_model.remove_state(bytes_buffer._id)
                logger.info("Cleaned up resources after WebSocket disconnect")

        uvicorn.run(app, host=self.host, port=self.port, ws="wsproto")

    async def vad_task(
        self,
        ws: WebSocket,
        span_queue: Queue,
        bytes_buffer: AudioBytesStream,
        vad_done_event: Event,
    ):
        logger.info("start vad task")
        while True:
            try:
                raw_data = await ws.receive()
            except Exception as e:
                logger.error(f"Error receiving data: {e}")
                vad_done_event.set()
                break
            bytes_data = raw_data.get("bytes", None)
            if bytes_data is None:
                logger.warning("Received data is None")
                continue
            chunks: List[AudioChunk] = bytes_buffer.push(bytes_data)
            for chunk in chunks:
                for segment_chunk in self.vad_model.detect_chunk(chunk=chunk):
                    segment_chunk: AudioChunk
                    if segment_chunk.vad_state != "segment_mid":
                        await self.send_vad_response(
                            "",
                            ws,
                            segment_chunk.vad_state,
                            start_time=segment_chunk.start_ms,
                            end_time=segment_chunk.end_ms,
                        )
                    await span_queue.put(segment_chunk)

    async def asr_task(self, ws: WebSocket, span_queue: Queue, vad_is_done: Event):
        final_text = ""
        segment: AudioChunkList = AudioChunkList()
        while not vad_is_done.is_set():
            span: AudioChunk = await span_queue.get()
            segment.append(span)
            is_last = span.vad_state == "segment_end"
            if is_last:
                if self.offline_asr_model is None:
                    for token in self.asr_model.transcribe_chunk(chunk=span):
                        final_text += token.text
                        await self.send_asr_response(
                            final_text, ws, "interim_transcript"
                        )
                    segment_chunk: AudioChunk = segment.concat_to_chunk()
                    if (
                        len(final_text) > 0
                        and segment_chunk.duration_ms >= self.min_segment_duration_ms
                    ):
                        if self.punc_model is not None:
                            restored_text = self.punc_model.restore(final_text).text
                            await self.send_asr_response(
                                restored_text,
                                ws,
                                "final_transcript",
                                start_time=segment_chunk.start_ms,
                                end_time=segment_chunk.end_ms,
                                raw_text=final_text,
                            )
                        else:
                            await self.send_asr_response(
                                final_text,
                                ws,
                                "final_transcript",
                                start_time=segment_chunk.start_ms,
                                end_time=segment_chunk.end_ms,
                            )
                else:
                    segment_chunk: AudioChunk = segment.concat_to_chunk()
                    segment_data = segment_chunk.waveform.data
                    final_text = self.offline_asr_model.transcribe(
                        batch=[segment_data]
                    )[0].text
                    if (
                        len(final_text) > 0
                        and segment_chunk.duration_ms >= self.min_segment_duration_ms
                    ):
                        if self.punc_model is not None:
                            restored_text = self.punc_model.restore(final_text).text
                        else:
                            restored_text = final_text
                        await self.send_asr_response(
                            restored_text,
                            ws,
                            "final_transcript",
                            start_time=segment_chunk.start_ms,
                            end_time=segment_chunk.end_ms,
                            raw_text=final_text,
                        )
                    self.asr_model.remove_state(
                        segment_chunk.stream_id
                    )  # clear state, because the segment is ended

                final_text = ""
                segment = AudioChunkList()
            else:
                for token in self.asr_model.transcribe_chunk(chunk=span):
                    token: AudioToken
                    final_text += token.text
                    await self.send_asr_response(final_text, ws, "interim_transcript")
            span_queue.task_done()

    async def send_asr_response(
        self,
        text: str,
        ws: WebSocket,
        state: str,
        start_time: float = None,
        end_time: float = None,
        raw_text: str = None,
    ):
        if raw_text is not None:
            if len(raw_text) == 1 and text in ["嗯", "啊"]:
                return
        if len(text) > 0:
            response = TranscriptionResponse(
                data=AudioChunkResponse(
                    text=text, state=state, start_time=start_time, end_time=end_time
                )
            )
            await ws.send_json(response.model_dump())
            logger.info(f"asr state: {state}, text: {text}")

    async def send_vad_response(
        self,
        text: str,
        ws: WebSocket,
        state: str,
        start_time: float = None,
        end_time: float = None,
    ):
        response = TranscriptionResponse(
            data=AudioChunkResponse(
                text=text, state=state, start_time=start_time, end_time=end_time
            )
        )
        await ws.send_json(response.model_dump())
        logger.info(f"vad state: {state}")

    @property
    def min_segment_duration_ms(self) -> int:
        return self.min_speech_duration_ms + self.vad_end_silence_ms



from typing import List

from joblib import Parallel, delayed
from pydantic import Field
from typing_extensions import Self

from fasr.components.base import BaseComponent
from fasr.config import registry
from fasr.data import AudioChannel, AudioList
from fasr.model import LIDModel


@registry.components.register("identifier")
class LanguageIdentifier(BaseComponent):
    name: str = "identifier"
    description: str = "Identify spoken language of audio channels"
    input_tags: List[str] = ["Audio.channels"]
    output_tags: List[str] = ["AudioChannel.language"]

    model: LIDModel | None = Field(
        None, description="The LID model to use for language identification"
    )
    num_threads: int | None = Field(
        None, description="The number of threads to use for parallel processing"
    )

    def predict(self, audios: AudioList) -> AudioList:
        all_channels = [channel for audio in audios for channel in audio.channels]
        batch_size = max(1, len(all_channels) // self.num_threads)
        _ = Parallel(n_jobs=self.num_threads, prefer="threads", batch_size=batch_size)(
            delayed(self.predict_channel)(channel) for channel in all_channels
        )
        return audios

    def predict_channel(self, channel: AudioChannel) -> AudioChannel:
        language = self.model.identify(waveform=channel.waveform)
        channel.language = language
        return channel

    def setup(
        self,
        model: str,
        num_threads: int = 1,
        **kwargs,
    ) -> Self:
        self.load_model_plugin(group="fasr.lid_models", model=model)
        model: LIDModel = registry.lid_models.get(model)()
        self.model = model.from_checkpoint(**kwargs)
        self.num_threads = num_threads
        return self



from typing import Iterable, List

from joblib import Parallel, delayed
from pydantic import validate_call
from typing_extensions import Self

from fasr.config import registry
from fasr.data.audio import (
    Audio,
    AudioChannel,
    AudioList,
    AudioSpan,
    AudioSpanList,
    AudioToken,
    AudioTokenList,
)
from fasr.model import ASRModel

from .base import BaseComponent


@registry.components.register("recognizer")
class SpeechRecognizer(BaseComponent):
    name: str = "recognizer"
    description: str = "Recognize speech from audio channels"
    input_tags: List[str] = ["Audio.channels"]
    output_tags: List[str] = ["AudioChannel.tokens"]
    model: ASRModel | None = None
    num_threads: int = 1
    batch_size_s: int | None = None
    hotwords: frozenset[str] | None = None

    def predict(self, audios: AudioList) -> AudioList:
        if self.hotwords is not None:
            for audio in audios:
                audio: Audio
                if audio.hotwords is None:
                    audio.set_hotwords(self.hotwords)
        groups = audios.group_by_hotwords()
        _ = Parallel(
            n_jobs=self.num_threads, prefer="threads", pre_dispatch="1 * n_jobs"
        )(
            delayed(self.predict_step)(batch_segments, list(hotwords))
            for hotwords, group_audios in groups.items()
            for batch_segments in self.batch_audio_segments(audios=group_audios)
        )
        return audios

    def predict_step(
        self, batch_segments: List[AudioSpan | AudioChannel], hotwords: List[str] = []
    ) -> List[AudioSpan | AudioChannel]:
        batch_waveforms = [seg.waveform for seg in batch_segments]
        if len(batch_waveforms) == 0:
            return batch_segments
        sample_rate = batch_waveforms[0].sample_rate  # 一个batch的音频片段采样率相同
        batch_data = [waveform.data for waveform in batch_waveforms]
        batch_tokens = self.model.transcribe(
            batch=batch_data, sample_rate=sample_rate, hotwords=hotwords
        )
        for seg, tokens in zip(batch_segments, batch_tokens):
            if not isinstance(seg, AudioChannel):
                if seg.start_ms is not None and seg.end_ms is not None:
                    for token in tokens:
                        token: AudioToken
                        if (
                            token.start_ms and token.end_ms
                        ):  # not all asr model predict timestamp
                            token.start_ms += seg.start_ms
                            token.end_ms += seg.start_ms
            seg.tokens = AudioTokenList(docs=tokens)
        return batch_segments

    def batch_audio_segments(self, audios: AudioList) -> Iterable[AudioSpanList]:
        """将音频片段组成批次。
        步骤：
        - 1. 将音频片段按照时长排序。
        - 2. 将音频片段按照时长分组，每组时长不超过batch_size_s。
        """
        all_segments = []
        for audio in audios:
            if not audio.is_bad:
                for channel in audio.channels:
                    if channel.segments is None:  # 兼容没有vad模型的情况
                        all_segments.append(channel)
                    else:
                        for seg in channel.segments:
                            all_segments.append(seg)
        return self.batch_segments(all_segments)

    def batch_segments(
        self, segments: Iterable[AudioSpan | AudioChannel]
    ) -> Iterable[List[AudioSpan | AudioChannel]]:
        """将音频片段组成动态批次（按时长贪心组批）。

        组批目标是让同一批次内的样本尽量“时长相近”，从而降低 padding 带来的浪费。

        具体规则：
        - **批次上限**：`batch_size_s` 表示一个批次的“等效总时长上限”（秒），会换算成
          毫秒 `batch_size_ms = batch_size_s * 1000`。
          当 `batch_size_s is None` 时，不按时长上限切分，直接把全部片段组成一个批次。
        - **排序**：先将 `segments` 收集成 list，并按 `duration_ms` 升序排序。
        - **贪心装箱**：遍历排序后的片段，维护当前批次的最大时长 `batch_max_duration_ms`。
          尝试把片段加入当前批次时，计算加入后的等效总时长：

            `current = max(batch_max_duration_ms, seg.duration_ms) * (len(batch) + 1)`

          这里的 `max_duration_ms * batch_size` 等价于“把批次内所有样本 padding 到同一最大时长”的
          计算量/时长近似。
        - **切批**：若 `current > batch_size_ms`，先 `yield` 当前批次，再以当前片段开启新批次；
          否则将片段加入当前批次并更新 `batch_max_duration_ms`。
        - **收尾**：遍历结束后若当前批次非空，会再 `yield` 一次。

        Returns:
            Iterable[List[AudioSpan | AudioChannel]]: 每次产出一个批次（list）。
        """
        segments = [seg for seg in segments]
        sorted_segments = self.sort_segments(segments)
        if self.batch_size_s is None:
            if len(sorted_segments) > 0:
                yield sorted_segments
            return

        batch_size_ms = self.batch_size_s * 1000
        batch: List[AudioSpan | AudioChannel] = []
        batch_max_duration_ms = 0
        for seg in sorted_segments:
            max_duration_ms = max(batch_max_duration_ms, seg.duration_ms)
            current_batch_duration_ms = max_duration_ms * (
                len(batch) + 1
            )  # 如果加入当前片段，是否超过batch_size_s
            if current_batch_duration_ms > batch_size_ms:
                yield batch
                batch = [seg]
                batch_max_duration_ms = seg.duration_ms
            else:
                batch.append(seg)
                batch_max_duration_ms = max_duration_ms
        if len(batch) > 0:
            yield batch

    def sort_segments(
        self, segments: List[AudioSpan | AudioChannel]
    ) -> List[AudioSpan | AudioChannel]:
        return sorted(segments, key=lambda x: x.duration_ms)

    @validate_call
    def setup(
        self,
        model: str,
        batch_size_s: int | None = None,
        num_threads: int = 1,
        hotwords: Iterable[str] | None = None,
        **kwargs,
    ) -> Self:
        self.load_model_plugin(group="fasr.asr_models", model=model)
        model: ASRModel = registry.asr_models.get(model)()
        self.model = model.from_checkpoint(**kwargs)
        self.num_threads = num_threads
        self.batch_size_s = batch_size_s
        self.hotwords = frozenset(hotwords) if hotwords is not None else None
        return self
